//////////////////////////////////////////////////////////////////////////////
//
//  Xilinx, Inc. 2002                 www.xilinx.com
//
//  XAPP 268 - Automatic Phase Alignment
//
//////////////////////////////////////////////////////////////////////////////
//
//  File name :       phase_control_full_range.v
//
//  Description :     Automatic phase controller, working over full range of
//                    DCM (-255 to +255)
//
//  Date - revision : 15th May 2003 - v 2.1
//
//  Author :          NJS
//
//  Contact : e-mail  hotline@xilinx.com
//            phone   + 9'b000000001 800 255 7778
//
//  Disclaimer: LIMITED WARRANTY AND DISCLAMER. These designs are
//              provided to you "as is". Xilinx and its licensors make, and you
//              receive no warranties or conditions, express, implied,
//              statutory or otherwise, and Xilinx specifically disclaims any
//              implied warranties of merchantability, non-infringement, or
//              fitness for a particular purpose. Xilinx does not warrant that
//              the functions contained in these designs will meet your
//              requirements, or that the operation of these designs will be
//              uninterrupted or error free, or that defects in the Designs
//              will be corrected. Furthermore, Xilinx does not warrant or
//              make any representations regarding use or the results of the
//              use of the designs in terms of correctness, accuracy,
//              reliability, or otherwise.
//
//              LIMITATION OF LIABILITY. In no event will Xilinx or its
//              licensors be liable for any loss of data, lost profits, cost
//              or procurement of substitute goods or services, or for any
//              special, incidental, consequential, or indirect damages
//              arising from the use or operation of the designs or
//              accompanying documentation, however caused and on any theory
//              of liability. This limitation will apply even if Xilinx
//              has been advised of the possibility of such damage. This
//              limitation shall apply not-withstanding the failure of the
//              essential purpose of any limited remedies herein.
//
//  Copyright  2002 Xilinx, Inc.
//  All rights reserved
//
//////////////////////////////////////////////////////////////////////////////
//
`timescale 1 ps / 1ps

module phase_control_full_range (
	enable,					// enable input, normally high, but allows function to be bypassed to speed up simulation when low
	monclkin,				// clock to be monitored
	rxclkin,				// clock from dcm
	ctlclk,					// phase_control clock
	lockedin,				// dcm locked in
	rst,					// reset
	psdone,					// high when dcm has changed phase
	forceup,				// forces phase value up when high
	forcedown,				// forces phase value down when high
	dcm_div_2_used,				// normally 1'b0, set to 1'b1 if DCM divide_by_2 option
						// is used (SDR designs)
	k,					// constant to be modified in the case where the data
						// and clock do not arrive at the part exactly aligned
						// permissable values 0 to 255 (0 to 360 degrees),
						// set default to "00000000"
	lockedout,				// locked out
	psincdec,				// move phase up (not)down
	psen,					// phase shift enable
	psclk,					// phase shift clock
	binout,					// binary current phase  value (0 to 511)
	bcdout);				// 12 downto 0  : BCD current phase value ( -255 to +255)
						// 26 downto 14 : BCD for ps0
						// 39 downto 27 : BCD for ps1

input 		enable, monclkin, rxclkin, ctlclk, lockedin, rst, psdone, forceup, forcedown, dcm_div_2_used ;
input 	[6:0] 	k ;
output 		lockedout, psincdec, psen, psclk ;
output 	[39:0] 	bcdout ;
output 	[8:0] 	binout ;
reg 		lockedout ;

// wires

wire 		low  ;			// logic low
wire 		high  ;			// logic high
wire 		forceup_int  ;		// internal force up wire
wire 		forcedown_int  ;	// internal force down wire
wire 		psdone_int  ;		// internal psdone wire
wire 		lockedin_int  ;		// internal lockedin wire
reg 		psen_int  ;		// internal psen wire
reg 		psincdec_int  ;		// internal psincdec wire
wire 		cal1,cal0  ;		// internal FF outputs
reg 		cal2  ;			// internal FF outputs
reg 		oldcal2 ;		// delayed cal2
wire		cal2a ;			// registered cal2
reg		calvalid ;		// indicates filter has finished
reg 		action ;		// indicates action (inc or dec) in progress
reg 		gt ;			// indicates when ps3 is greater than ps1
reg 	[2:0] 	state ;			// state counter
wire 	[12:0] 	bcdint  ;		// internal phase value in BCD
reg 	[8:0] 	ps0  ;			// ps0 found
reg 	[8:0] 	ps1  ;			// ps1 found
reg 	[12:0] 	ps0bcd  ;		// ps0 found in BCD
reg 	[12:0] 	ps1bcd  ;		// ps1 found in BCD
reg 	[8:0] 	ps3  ;			// calclated ps3
wire 	[9:0] 	ps3a  ;			// used to calculate ps3
wire 	[9:0] 	ps3b  ;			// used to calculate ps3
wire 	[9:0] 	ps3c  ;			// used to calculate ps3
reg 	[3:0] 	hyst  ;			// hysteresis counter
reg 	[23:0] 	bigcounter  ;		// long counter for time between increments/decrements when forcing up or down
reg	[3:0]	count ;	 		// used to build in some hysteresis
reg	[8:0]	int ;	 		// binary counter

assign low = 1'b0 ;
assign high = 1'b1 ;
assign psen = psen_int ;
assign psincdec = psincdec_int ;
assign bcdout = {ps1bcd, ps0bcd, low, bcdint} ;
assign binout = int ;
assign psclk = ctlclk ;
assign ps3a = {low, ps0} + {low, ps1} ;
assign ps3b = {low, ps3a[8:1]} - {3'b000, k} ;
assign ps3c = {low, ps3a[8:1]} + 10'b0100000000 - {3'b000, k} ;


FD fd_regcal0 (.C(rxclkin),    .D(monclkin), .Q(cal0))			// IOB risng edge reg
/*synthesis iob = "true" */						// Synplicity attribute
/* synthesis attribute IOB of fd_regcal0 "TRUE */;			// XST attribute
FD fd_regcal1 (.C(rxclkin),    .D(cal0),     .Q(cal1))
/*synthesis rloc = "x0y0" */
/* synthesis attribute RLOC of fd_regcal1 X0Y0 */;
FD fd_regcal2 (.C(ctlclk),     .D(cal1),     .Q(cal2a))
/*synthesis rloc = "x1y0" */
/* synthesis attribute RLOC of fd_regcal2 X1Y0 */;			// retime to ctlclk

FD fd_retimepsdone (.C(ctlclk),     .D(psdone),   .Q(psdone_int));	// retime psdone to ctlclk
FD fd_retimelocked (.C(ctlclk),     .D(lockedin), .Q(lockedin_int));	// retime lockedin to ctlclk

FD fd_retime_fup (.C(ctlclk),     .D(forceup),   .Q(forceup_int));	// retime forceup to ctlclk
FD fd_retime_fdn (.C(ctlclk),     .D(forcedown), .Q(forcedown_int));	// retime forcedown to ctlclk

bcd35_full_range count0 (
	.clk(ctlclk),
	.rst(rst),
	.enablein(psdone_int),
	.up(psincdec_int),
	.bcdvec(bcdint));

// 9 bit counter process
always @ (posedge ctlclk or posedge rst)
begin
if (rst) begin
	count <= 4'b1000 ;
	cal2 <= 1'b0 ;
	calvalid <= 1'b0 ;
end
else begin
	if (cal2a == 1'b1) begin
		if (count < 4'b1111) begin
			count <= count + 4'b0001 ;
		end
	end
	else begin
		if (count > 4'b0000) begin
			count <= count - 4'b0001 ;
		end
	end
	if (count == 4'b1111) begin
		cal2 <= 1'b1 ;
		calvalid <= 1'b1 ;
	end
	else if (count == 4'b0000) begin
		cal2 <= 1'b0 ;
		calvalid <= 1'b1 ;
	end
	else begin
		calvalid <= 1'b0 ;
	end
end
end
			
// state machine
always @ (posedge ctlclk or posedge rst)
begin
if (rst) begin
	bigcounter <= 24'b000000000000000000000000 ;
	ps0bcd <= 13'b0000000000000 ;
	ps1bcd <= 13'b0000000000000 ;
	psincdec_int <= 1'b1 ;
	psen_int <= 1'b0 ;
	ps0 <= 9'b000000000 ;
	ps1 <= 9'b000000000 ;
	ps3 <= 9'b000000000 ;
	hyst <= 4'b0000 ;
	action <= 1'b0 ;
	gt <= 1'b0 ;
	state <= 0 ;
	lockedout <= 1'b0 ;
	oldcal2 <= 1'b0 ;
	int <= 9'b000000000 ;
end
else begin
	if (k[6] == 1'b1) begin							// K > 64 ??
		gt <= 1'b1 ;
		ps3 <= ps3c[8:0] ;
	end
	else begin
		gt <= 1'b0 ;
		ps3 <= ps3b[8:0] ;
	end
	bigcounter <= bigcounter + 24'b000000000000000000000001 ;
	if (psdone_int) begin
		action <= 1'b0 ;
	end
	if (lockedin_int && ~action && calvalid) begin
		oldcal2 <= cal2 ;
		if (enable == 1'b0) begin
			lockedout <= 1'b1 ;
			state <= 5 ;
		end
		else if (state == 0) begin
			state <= 1 ;
		end
		else if (state == 1) begin
			psen_int <= 1'b1 ;					// increment
			action <= 1'b1 ;
			int <= int + 9'b000000001 ;
			if ((~dcm_div_2_used && cal2 == ~oldcal2 && int > 9'b001000000) ||		// find an edge (rising if SDR)
			   (dcm_div_2_used && cal2 == ~oldcal2 && int > 9'b001000000 && cal2==1'b1 )) begin
				state <= 2 ;
				hyst <= 4'b0000 ;
				ps0bcd <= bcdint ;
				ps0 <= int ;					// save value
			end
		end
		else if (state == 2) begin					// build in some hysteresis
			psen_int <= 1'b1 ;
			action <= 1'b1 ;
			int <= int + 9'b000000001 ;
			hyst <= hyst + 4'b0001 ;
			if (hyst == 4'b1111) begin
				state <= 3 ;
			end
		end
		else if (state == 3) begin
			psen_int <= 1'b1 ;					// increment
			action <= 1'b1 ;
			int <= int + 9'b000000001 ;
			if ((~dcm_div_2_used && cal2 == ~oldcal2) ||
			   (dcm_div_2_used && cal2 == ~oldcal2 && cal2 ) ||
			   (int == 9'b111111110)) begin		// find another edge (risng for SDR) or an overflow condition
				state <= 4 ;
				ps1 <= int ;
				ps1bcd <= bcdint ;
			end
		end
		else if (state == 4) begin
			psincdec_int <= gt ;
			psen_int <= 1'b1 ;
			action <= 1'b1 ;
			if (gt == 1'b1) begin
				int <= int + 9'b000000001 ;
			end
			else begin
				int <= int - 9'b000000001 ;
			end
			if (int == ps3) begin
				state <= 5 ;
				lockedout <= 1'b1 ;
			end
		end
		else if (state == 5 && bigcounter == 24'b111111111111111111111111) begin
			if (forceup_int) begin
				psincdec_int <= 1'b1 ;
				psen_int <= 1'b1 ;
				action <= 1'b1 ;
				int <= int + 9'b000000001 ;
			end
			else if (forcedown_int) begin
				psincdec_int <= 1'b0 ;
				psen_int <= 1'b1 ;
				action <= 1'b1 ;
				int <= int - 9'b000000001 ;
			end
		end
	end
	else begin
		psen_int <= 1'b0 ;
	end
end
end
endmodule