------------------------------------------------------------------------------
--
--  Xilinx, Inc. 2002                 www.xilinx.com
--
--  XAPP 268 - Automatic Phase Alignment
--
------------------------------------------------------------------------------
--
--  File name :       phase_control_full_range.vhd
--
--  Description :     Automatic phase controller, working over full range of
--                    DCM (-255 to +255)
--
--  Date - revision : 20th May 2003 - v 2.1
--
--  Author :          NJS
--
--  Contact : e-mail  hotline@xilinx.com
--            phone   + 1 800 255 7778
--
--  Disclaimer: LIMITED WARRANTY AND DISCLAMER. These designs are
--              provided to you "as is". Xilinx and its licensors make, and you
--              receive no warranties or conditions, express, implied,
--              statutory or otherwise, and Xilinx specifically disclaims any
--              implied warranties of merchantability, non-infringement, or
--              fitness for a particular purpose. Xilinx does not warrant that
--              the functions contained in these designs will meet your
--              requirements, or that the operation of these designs will be
--              uninterrupted or error free, or that defects in the Designs
--              will be corrected. Furthermore, Xilinx does not warrant or
--              make any representations regarding use or the results of the
--              use of the designs in terms of correctness, accuracy,
--              reliability, or otherwise.
--
--              LIMITATION OF LIABILITY. In no event will Xilinx or its
--              licensors be liable for any loss of data, lost profits, cost
--              or procurement of substitute goods or services, or for any
--              special, incidental, consequential, or indirect damages
--              arising from the use or operation of the designs or
--              accompanying documentation, however caused and on any theory
--              of liability. This limitation will apply even if Xilinx
--              has been advised of the possibility of such damage. This
--              limitation shall apply not-withstanding the failure of the
--              essential purpose of any limited remedies herein.
--
--  Copyright  2002 Xilinx, Inc.
--  All rights reserved
--
------------------------------------------------------------------------------
--
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all ;

--pragma translate_off
library unisim ;
use unisim.vcomponents.all ;
--pragma translate_on

entity phase_control_full_range is port (
	enable			: in std_logic;				-- enable input, normally high, but allows function to be bypassed to speed up simulation when low
	monclkin		: in std_logic;				-- clock to be monitored
	rxclkin			: in std_logic;				-- clock from dcm
	ctlclk			: in std_logic;				-- phase_control_full_range clock
	lockedin		: in std_logic;				-- dcm locked in
	rst			: in std_logic;				-- reset
	psdone			: in std_logic;				-- high when dcm has changed phase
	forceup			: in std_logic;				-- forces phase value up when high
	forcedown		: in std_logic;				-- forces phase value down when high
	dcm_div_2_used		: in std_logic;				-- normally '0', set to '1' if DCM divide_by_2 option
									-- is used (SDR designs)
	k 			: in std_logic_vector(6 downto 0);	-- constant to be modified in the case where the data
									-- and clock do not arrive at the part exactly aligned
									-- permissable values 0 to 127 (0 to 180 degrees),
									-- set default to "0000000"
	lockedout		: out std_logic;			-- locked out
	psincdec		: out std_logic ;			-- move phase up (not)down
	psen			: out std_logic ;			-- phase shift enable
	psclk			: out std_logic ;			-- phase shift clock
	binout			: out std_logic_vector(8 downto 0);	-- binary current phase  value (0 to 511)
	bcdout			: out std_logic_vector(39 downto 0));	-- 12 downto 0  : BCD current phase value ( -255 to +255)
									-- 26 downto 14 : BCD for ps0
									-- 39 downto 27 : BCD for ps1
end phase_control_full_range;

architecture arch_phase_control_full_range of phase_control_full_range is

-- components

-- three and a half decade BCD counter
component bcd35_full_range port (
	clk		: in std_logic;				-- clock
	rst		: in std_logic;				-- reset
	enablein	: in std_logic;				-- enable in
	up		: in std_logic;				-- count up
	bcdvec		: out std_logic_vector(12 downto 0));	-- 3 1/2 decade output
end component ;
-- FF primitive
component fd
--pragma translate_off
		generic(TimingChecksOn: Boolean := FALSE) ;
--pragma translate_on
		port(c, d : in std_logic; q : out std_logic); end component ;

-- signals

signal low 		: std_logic ;				-- logic low
signal high 		: std_logic ;				-- logic high
signal forceup_int 	: std_logic ;				-- internal force up signal
signal forcedown_int 	: std_logic ;				-- internal force down signal
signal psdone_int 	: std_logic ;				-- internal psdone signal
signal lockedin_int 	: std_logic ;				-- internal lockedin signal
signal psen_int 	: std_logic ;				-- internal psen signal
signal psincdec_int 	: std_logic ;				-- internal psincdec signal
signal cal2, cal1, cal0 : std_logic ;				-- internal FF outputs
signal oldcal2, cal2a	: std_logic ;				-- registered cal2
signal calvalid		: std_logic ;				-- indicates filter has finished
signal action 		: std_logic ;				-- indicates action (inc or dec) in progress
signal gt 		: std_logic ;				-- indicates when ps3 is greater than ps1
signal state 		: integer range 0 to 7 ;		-- state counter
signal bcdint 		: std_logic_vector(12 downto 0) ;	-- internal phase value in BCD
signal int 		: std_logic_vector(8 downto 0) ;	-- internal phase value in binary
signal ps0 		: std_logic_vector(8 downto 0) ;	-- ps0 found
signal ps1 		: std_logic_vector(8 downto 0) ;	-- ps1 found
signal ps0bcd 		: std_logic_vector(12 downto 0) ;	-- ps0 found in BCD
signal ps1bcd 		: std_logic_vector(12 downto 0) ;	-- ps1 found in BCD
signal ps3 		: std_logic_vector(8 downto 0) ;	-- calclated ps3
signal ps3a 		: std_logic_vector(9 downto 0) ;	-- used to calculate ps3
signal ps3b 		: std_logic_vector(9 downto 0) ;	-- used to calculate ps3
signal ps3c 		: std_logic_vector(9 downto 0) ;	-- used to calculate ps3
signal hyst 		: std_logic_vector(3 downto 0) ;	-- hysteresis counter
signal bigcounter 	: std_logic_vector(23 downto 0) ;	-- long counter for time between increments/decrements when forcing up or down
signal count	 	: std_logic_vector(3 downto 0) ;	-- used to build in some hysteresis

-- attributes

attribute rloc		: string ;
attribute iob		: string ;

attribute iob of fd_regcal0	: label is "true" ;		-- forces clock receiver FFs into IOBs
attribute rloc of fd_regcal1	: label is "x0y0" ;
attribute rloc of fd_regcal2	: label is "x1y0" ;

begin

low 	<= '0' ;
high 	<= '1' ;
psen <= psen_int ;
psincdec <= psincdec_int ;
bcdout <= ps1bcd & ps0bcd & '0' & bcdint ;
psclk <= ctlclk ;
binout <= int ;
ps3a <= ('0' & ps0) + ('0' & ps1) ;
ps3b <= ('0' & ps3a(8 downto 1)) - ("000" & k) ;
ps3c <= (('0' & ps3a(8 downto 1)) + "0100000000") - ("000" & k) ;

fd_regcal0 : fd  port map (c => rxclkin, d => monclkin,  q => cal0);			-- IOB risng edge reg
fd_regcal1 : fd  port map (c => rxclkin, d => cal0,      q => cal1);
fd_regcal2 : fd  port map (c => ctlclk,  d => cal1,      q => cal2a);			-- retime to ctlclk

fd_retimepsdone : fd  port map (c => ctlclk,  d => psdone,    q => psdone_int);		-- retime psdone to ctlclk
fd_retimelocked : fd  port map (c => ctlclk,  d => lockedin,  q => lockedin_int);	-- retime lockedin to ctlclk

fd_retime_fup   : fd  port map (c => ctlclk,  d => forceup,   q => forceup_int);	-- retime forceup to ctlclk
fd_retime_fdn   : fd  port map (c => ctlclk,  d => forcedown, q => forcedown_int);	-- retime forcedown to ctlclk

count0 : bcd35_full_range port map (
	clk		=> ctlclk,
	rst		=> rst,
	enablein	=> psdone_int,
	up		=> psincdec_int,
	bcdvec		=> bcdint);

process (ctlclk, rst)
begin
if rst = '1' then
	count <= "1000" ;
	cal2 <= '0' ;
	calvalid <= '0' ;
elsif ctlclk'event and ctlclk = '1' then
	if cal2a = '1' then
		if count < "1111" then
			count <= count + 1 ;
		end if ;
	else
		if count > "0000" then
			count <= count - 1 ;
		end if ;
	end if ;
	if count = "1111" then
		cal2 <= '1' ;
		calvalid <= '1' ;
	elsif count = "0000" then
		cal2 <= '0' ;
		calvalid <= '1' ;
	else
		calvalid <= '0' ;
	end if ;
end if ;
end process ;

process (ctlclk, rst)
begin
if rst = '1' then
	bigcounter <= (others => '0') ;
	ps0bcd <= (others => '0') ;
	ps1bcd <= (others => '0') ;
	psincdec_int <= '1' ;
	psen_int <= '0' ;
	ps0 <= (others => '0') ;
	ps1 <= (others => '0') ;
	ps3 <= (others => '0') ;
	hyst <= (others => '0') ;
	action <= '0' ;
	state <= 0 ;
	lockedout <= '0' ;
	gt <= '0' ;
	oldcal2 <= '0' ;
	int <= (others => '0') ;
elsif ctlclk'event and ctlclk = '1' then
	if k(6) = '1' then							-- K > 64 ??
		gt <= '1' ;
		ps3 <= ps3c(8 downto 0) ;
	else
		gt <= '0' ;
		ps3 <= ps3b(8 downto 0) ;
	end if ;
	bigcounter <= bigcounter + 1 ;
	if psdone_int = '1' then
		action <= '0' ;
	end if ;
	if lockedin_int = '1' and action = '0' and calvalid = '1' then
		oldcal2 <= cal2 ;
		if enable = '0' then
			lockedout <= '1' ;
			state <= 5 ;
		elsif state = 0 then						-- dummy state
			state <= 1 ;
		elsif state = 1 then
			psen_int <= '1' ;					-- increment
			action <= '1' ;
			int <= int + 1 ;
--			if (dcm_div_2_used = '0' and cal2 /= oldcal2 ) or	-- find an edge (rising if SDR)
			if (dcm_div_2_used = '0' and cal2 /= oldcal2 and int > "001000000" ) or	-- find an edge (rising if SDR)
			   (dcm_div_2_used = '1' and cal2 /= oldcal2 and cal2 = '1' and int > "001000000" ) then
				state <= 2 ;
				hyst <= "0000" ;
				ps0bcd <= bcdint ;
				ps0 <= int ;					-- save value
			end if ;
		elsif state = 2 then						-- build in some hysteresis
			psen_int <= '1' ;
			action <= '1' ;
			int <= int + 1 ;
			hyst <= hyst + 1 ;
			if hyst = "1111" then
				state <= 3 ;
			end if ;
		elsif state = 3 then
			psen_int <= '1' ;					-- increment
			action <= '1' ;
			int <= int + 1 ;
			if (dcm_div_2_used = '0' and cal2 /= oldcal2) or
			   (dcm_div_2_used = '1' and cal2 /= oldcal2 and cal2 = '1' ) or
			   (int = "111111110") then		-- find another edge (risng for SDR) or an overflow condition
				state <= 4 ;
				ps1 <= int ;
				ps1bcd <= bcdint ;
			end if ;
		elsif state = 4 then
			psincdec_int <= gt ;
			psen_int <= '1' ;
			action <= '1' ;
			if gt = '1' then
				int <= int + 1 ;
			else
				int <= int - 1 ;
			end if ;
			if int = ps3 then
				state <= 5 ;
				lockedout <= '1' ;
			end if ;
		elsif state = 5 and bigcounter = "111111111111111111111111" then
			if forceup_int = '1' then
				psincdec_int <= '1' ;
				psen_int <= '1' ;
				action <= '1' ;
				int <= int + 1 ;
			elsif forcedown_int = '1' then
				psincdec_int <= '0' ;
				psen_int <= '1' ;
				action <= '1' ;
				int <= int - 1 ;
			end if ;
		end if ;
	else
		psen_int <= '0' ;
	end if ;
end if ;
end process ;
end arch_phase_control_full_range;


